# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import inspect
import typing
from base64 import b64encode

from servicex.configuration import Configuration
from servicex.query import Query
from servicex.models import ResultFormat
from servicex.query_cache import QueryCache
from servicex.servicex_adapter import ServiceXAdapter
from servicex.types import DID


class PythonQuery(Query):

    def __init__(self, dataset_identifier: DID,
                 sx_adapter: ServiceXAdapter = None,
                 title: str = "ServiceX Client",
                 codegen: str = None,
                 config: Configuration = None,
                 query_cache: QueryCache = None,
                 result_format: typing.Optional[ResultFormat] = None,
                 ignore_cache: bool = False
                 ):
        super().__init__(dataset_identifier=dataset_identifier,
                         title=title,
                         codegen=codegen,
                         sx_adapter=sx_adapter,
                         config=config,
                         query_cache=query_cache,
                         result_format=result_format,
                         ignore_cache=ignore_cache)

        self.python_function = None

    def with_uproot_function(self, f: typing.Union[str, typing.Callable]) -> Query:
        self.python_function = f
        return self

    def generate_selection_string(self) -> str:
        if not self.python_function:
            raise ValueError("You must provide a python function using with_uproot_function")

        if isinstance(self.python_function, str):
            return b64encode(self.python_function.encode("utf-8")).decode("utf-8")
        else:
            return b64encode(inspect.getsource(self.python_function)
                             .encode("utf-8"))\
                .decode("utf-8")



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import math
import os
import tempfile
from unittest.mock import patch

import httpx
import pytest
from pytest_asyncio import fixture

from servicex.models import TransformRequest, ResultDestination, ResultFormat
from servicex.servicex_adapter import ServiceXAdapter, AuthorizationError


@fixture
def servicex():
    return ServiceXAdapter("https://servicex.org")


def test_result_formats():
    """
    This test is just to make sure the enum string representations match the values
    expected by the service. Compare this to the json parser in
    servicex.resources.transformation.submit.SubmitTransformationRequest.make_api
    """
    assert ResultFormat.parquet == "parquet"
    assert ResultFormat.root == "root-file"


@pytest.mark.asyncio
@patch('servicex.servicex_adapter.httpx.AsyncClient.get')
async def test_get_transforms(get, servicex, transform_status_response):
    get.return_value = httpx.Response(200, json=transform_status_response)
    t = await servicex.get_transforms()
    assert len(t) == 1
    assert t[0].request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"
    get.assert_called_with(url='https://servicex.org/servicex/transformation', headers={})


@pytest.mark.asyncio
@patch('servicex.servicex_adapter.httpx.AsyncClient.get')
async def test_get_transforms_auth_error(get, servicex):
    with pytest.raises(AuthorizationError):
        get.return_value = httpx.Response(401)
        await servicex.get_transforms()


@pytest.mark.asyncio
@patch('servicex.servicex_adapter.httpx.AsyncClient.get')
@patch('servicex.servicex_adapter.httpx.AsyncClient.post')
@patch('servicex.servicex_adapter.jwt.decode')
async def test_get_transforms_wlcg_bearer_token(decode, post, get, servicex,
                                                transform_status_response):
    token_file = tempfile.NamedTemporaryFile(mode="w+t", delete=False)
    token_file.write("luckycharms")
    token_file.close()

    os.environ['BEARER_TOKEN_FILE'] = token_file.name

    get.return_value = httpx.Response(200, json=transform_status_response)
    decode.return_value = {'exp': math.inf}
    await servicex.get_transforms()

    # Try with an expired token
    with pytest.raises(AuthorizationError):
        decode.return_value = {'exp': 0.0}
        await servicex.get_transforms()

    os.remove(token_file.name)
    del os.environ['BEARER_TOKEN_FILE']


@pytest.mark.asyncio
@patch('servicex.servicex_adapter.httpx.AsyncClient.post')
@patch('servicex.servicex_adapter.httpx.AsyncClient.get')
async def test_get_transforms_with_refresh(get, post, transform_status_response):
    servicex = ServiceXAdapter(url="https://servicex.org", refresh_token="refrescas")
    post.return_value = httpx.Response(200, json={"access_token": "luckycharms"})
    get.return_value = httpx.Response(200, json=transform_status_response)
    await servicex.get_transforms()

    post.assert_called_with('https://servicex.org/token/refresh',
                            headers={'Authorization': 'Bearer refrescas'}, json=None)

    get.assert_called_with(url='https://servicex.org/servicex/transformation',
                           headers={'Authorization': 'Bearer luckycharms'})


@patch('servicex.servicex_adapter.httpx.Client.get')
def test_get_codegens(get, servicex):
    get.return_value = httpx.Response(200, json={
        "uproot": "http://uproot-codegen",
        "xaod": "http://xaod-codegen"
    })
    c = servicex.get_code_generators()
    assert len(c) == 2
    assert c["uproot"] == "http://uproot-codegen"


@pytest.mark.asyncio
@patch('servicex.servicex_adapter.httpx.AsyncClient.post')
async def test_submit(post, servicex):
    post.return_value = httpx.Response(200, json={"request_id": "123-456-789"})
    request = TransformRequest(
        title="Test submission",
        did="rucio://foo.bar",
        selection="(call EventDataset)",
        codegen="uproot",
        result_destination=ResultDestination.object_store,
        result_format=ResultFormat.parquet
    )
    result = await servicex.submit_transform(request)
    assert result == "123-456-789"


@pytest.mark.asyncio
@patch('servicex.servicex_adapter.httpx.AsyncClient.get')
async def test_get_transform_status(get, servicex, transform_status_response):
    get.return_value = httpx.Response(200, json=transform_status_response['requests'][0])
    result = await servicex.get_transform_status("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")
    assert result.request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"


@pytest.mark.asyncio
@patch('servicex.servicex_adapter.httpx.AsyncClient.get')
async def test_get_transform_status_auth_error(get, servicex):
    with pytest.raises(AuthorizationError):
        get.return_value = httpx.Response(401)
        await servicex.get_transform_status("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from unittest.mock import patch, MagicMock

from rich.progress import Progress

from servicex.expandable_progress import ExpandableProgress


@patch("servicex.expandable_progress.Progress", return_value=MagicMock(Progress))
def test_progress(mock_progress):
    with ExpandableProgress() as progress:
        assert progress.progress == mock_progress.return_value
        mock_progress.return_value.start.assert_called_once()
        assert progress.display_progress
    assert mock_progress.return_value.stop.call_count == 1


@patch("servicex.expandable_progress.Progress", return_value=MagicMock(Progress))
def test_overall_progress(mock_progress):
    with ExpandableProgress(overall_progress=True) as progress:
        assert progress.progress == mock_progress.return_value
        mock_progress.return_value.start.assert_called_once()
        assert progress.display_progress
    assert mock_progress.return_value.stop.call_count == 1


@patch("servicex.expandable_progress.Progress", return_value=MagicMock(Progress))
def test_overall_progress_mock(mock_progress):
    with ExpandableProgress(overall_progress=True) as progress:
        assert progress.progress == mock_progress.return_value
        mock_progress.return_value.start.assert_called_once()
        assert progress.display_progress
    assert mock_progress.return_value.stop.call_count == 1


def test_provided_progress(mocker):
    class MockedProgress(Progress):
        def __init__(self):
            self.start_call_count = 0
            self.stop_call_count = 0

        def start(self) -> None:
            self.start_call_count += 1

        def stop(self) -> None:
            self.stop_call_count += 1

    provided_progress = MockedProgress()
    provided_progress.start = mocker.Mock()
    provided_progress.stop = mocker.Mock()

    with ExpandableProgress(provided_progress=provided_progress) as progress:
        assert progress.progress == provided_progress
        assert provided_progress.start.call_count == 0
        assert progress.display_progress
    assert provided_progress.stop.call_count == 0


@patch("servicex.expandable_progress.Progress", return_value=MagicMock(Progress))
def test_no_progress(mock_progress):
    with ExpandableProgress(display_progress=False) as progress:
        assert not progress.progress
        mock_progress.return_value.assert_not_called()
        mock_progress.return_value.start.assert_not_called()
        assert not progress.display_progress
        assert not progress.progress
    assert mock_progress.return_value.stop.call_count == 0


def test_nested_expandable_progress():
    inner_progress = Progress()
    with ExpandableProgress(provided_progress=inner_progress) as progress:
        with ExpandableProgress(provided_progress=progress) as progress2:
            assert progress2.progress == progress.progress
            assert progress2.display_progress
            assert progress2.progress == progress.progress



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from pathlib import Path
from unittest.mock import AsyncMock

import pytest
from miniopy_async.datatypes import Object
from pytest_asyncio import fixture

from servicex.minio_adapter import MinioAdapter
from servicex.models import ResultFile


@fixture
def minio_adapter() -> MinioAdapter:
    return MinioAdapter("localhost", False, "access_key", "secret_key", "bucket")


@pytest.mark.asyncio
async def test_initialize_from_status(completed_status):
    minio = MinioAdapter.for_transform(completed_status)
    assert minio.minio._base_url.host == "minio.org:9000"
    assert minio.minio._provider._credentials.access_key == "miniouser"
    assert minio.minio._provider._credentials.secret_key == "secret"
    assert minio.bucket == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"


@pytest.mark.asyncio
async def test_list_bucket(minio_adapter):
    files = [ResultFile(filename="test.txt", size=10, extension="txt")]
    minio_adapter.minio.list_objects = AsyncMock(
        return_value=[
            Object(object_name=file.filename, size=file.size, bucket_name="bucket")
            for file in files
        ]
    )
    result = await minio_adapter.list_bucket()
    assert result == files
    minio_adapter.minio.list_objects.assert_called_with("bucket")


@pytest.mark.asyncio
async def test_download_file(minio_adapter):
    minio_adapter.minio.fget_object = AsyncMock(return_value="test.txt")
    result = await minio_adapter.download_file("test.txt", local_dir="/tmp/foo")
    assert str(result).endswith("test.txt")
    minio_adapter.minio.fget_object.assert_called_with(
        bucket_name="bucket", file_path="/tmp/foo/test.txt", object_name="test.txt"
    )


@pytest.mark.asyncio
async def test_download_bad_filename(minio_adapter):
    minio_adapter.minio.fget_object = AsyncMock(return_value="t::est.txt")
    result = await minio_adapter.download_file("t::est.txt", local_dir="/tmp/foo")
    assert str(result).endswith("t__est.txt")
    minio_adapter.minio.fget_object.assert_called_with(
        bucket_name="bucket", file_path="/tmp/foo/t__est.txt", object_name="t::est.txt"
    )


@pytest.mark.asyncio
async def test_download_short_filename_no_change(minio_adapter):
    minio_adapter.minio.fget_object = AsyncMock(return_value="test.txt")
    result = await minio_adapter.download_file(
        "test.txt", local_dir="/tmp/foo", shorten_filename=True
    )
    assert str(result).endswith("test.txt")
    minio_adapter.minio.fget_object.assert_called_with(
        bucket_name="bucket", file_path="/tmp/foo/test.txt", object_name="test.txt"
    )


@pytest.mark.asyncio
async def test_download_short_filename_change(minio_adapter):
    minio_adapter.minio.fget_object = AsyncMock(
        return_value="test12345678901234567890123456789012345678901234567898012345678901234567890.txt"  # noqa: E501
    )
    result = await minio_adapter.download_file(
        "test123456789012345678901234567890k12345678901234567898012345678901234567890.txt",
        local_dir="/tmp/foo",
        shorten_filename=True,
    )

    # Some of the filename should be left over still...
    assert str(result).endswith("01234567890.txt")

    # Make sure the length is right
    assert len(Path(result).name) == 60

    # Make sure we did the right call
    minio_adapter.minio.fget_object.assert_called_with(
        bucket_name="bucket",
        file_path="/tmp/foo/_7405534c58a3f71e5d01cdd2f59356bda6f50a06678901234567890.txt",
        object_name="test123456789012345678901234567890k12345678901234567898012345678901234567890.txt",  # noqa: E501
    )


@pytest.mark.asyncio
async def test_get_signed_url(minio_adapter):
    minio_adapter.minio.get_presigned_url = AsyncMock(
        return_value="https://pre-signed.me"
    )
    result = await minio_adapter.get_signed_url("test.txt")
    assert result == "https://pre-signed.me"
    minio_adapter.minio.get_presigned_url.assert_called_with(
        bucket_name="bucket", method="GET", object_name="test.txt"
    )



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from unittest.mock import AsyncMock

import pytest

from servicex import ResultFormat
from servicex.dataset_group import DatasetGroup


def test_set_result_format(mocker):
    ds1 = mocker.Mock()
    ds2 = mocker.Mock()
    group = DatasetGroup([ds1, ds2])
    group.set_result_format(ResultFormat.root)
    ds1.set_result_format.assert_called_once_with(ResultFormat.root)
    ds2.set_result_format.assert_called_once_with(ResultFormat.root)


@pytest.mark.asyncio
async def test_as_signed_urls(mocker, transformed_result):
    ds1 = mocker.Mock()
    ds1.as_signed_urls_async = AsyncMock(return_value=transformed_result)

    ds2 = mocker.Mock()
    ds2.as_signed_urls_async = AsyncMock(return_value=transformed_result.copy(
        update={"request_id": "98-765-432"}))

    group = DatasetGroup([ds1, ds2])
    results = await group.as_signed_urls_async()

    assert len(results) == 2
    assert results[0].request_id == "123-45-6789"
    assert results[1].request_id == "98-765-432"



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime

from pytest_asyncio import fixture
from servicex.python_dataset import PythonQuery
from servicex.models import (
    TransformRequest,
    ResultDestination,
    ResultFormat,
    TransformStatus,
    TransformedResults,
)

from servicex.dataset_identifier import FileListDataset
from servicex.minio_adapter import MinioAdapter

import pandas as pd
import os


@fixture
def transform_request() -> TransformRequest:
    return TransformRequest(
        title="Test submission",
        did="rucio://foo.bar",
        selection="(call EventDataset)",
        codegen="uproot",
        result_destination=ResultDestination.object_store,  # type: ignore
        result_format=ResultFormat.parquet,  # type: ignore
    )  # type: ignore


@fixture
def minio_adapter() -> MinioAdapter:
    return MinioAdapter("localhost", False, "access_key", "secret_key", "bucket")


@fixture
def python_dataset(dummy_parquet_file):
    did = FileListDataset(dummy_parquet_file)
    dataset = PythonQuery(
        title="Test submission",
        dataset_identifier=did,
        codegen="uproot",
        result_format=ResultFormat.parquet,  # type: ignore
    )  # type: ignore

    def foo():
        return

    dataset.with_uproot_function(foo)
    return dataset


@fixture
def transformed_result_python_dataset(dummy_parquet_file) -> TransformedResults:
    return TransformedResults(
        hash="289e90f6fe3780253af35c428b784ac22d3ee9200a7581b8f0a9bdcc5ae93479",
        title="Test submission",
        codegen="uproot",
        request_id="b8c508d0-ccf2-4deb-a1f7-65c839eebabf",
        submit_time=datetime.now(),
        data_dir="/foo/bar",
        file_list=[dummy_parquet_file],
        signed_url_list=[],
        files=1,
        result_format=ResultFormat.parquet,
    )


@fixture
def transform_status_response() -> dict:
    return {
        "requests": [
            {
                "request_id": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf",
                "did": "File List Provided in Request",
                "columns": None,
                "selection": "(Where (SelectMany (call EventDataset) (lambda (list e) (call (attr e 'Jets') 'AntiKt4EMTopoJets'))) (lambda (list j) (and (> (/ (call (attr j 'pt')) 1000) 20) (< (call abs (/ (call (attr j 'eta')) 1000)) 4.5))))",  # NOQA
                "tree-name": None,
                "image": "sslhep/servicex_func_adl_uproot_transformer:uproot4",
                "workers": None,
                "result-destination": "object-store",
                "result-format": "parquet",
                "workflow-name": "selection_codegen",
                "generated-code-cm": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf-generated-source",  # NOQA
                "status": "Submitted",
                "failure-info": None,
                "app-version": "develop",
                "code-gen-image": "sslhep/servicex_code_gen_func_adl_uproot:v1.2.0",
                "files": 1,
                "files-completed": 0,
                "files-failed": 0,
                "files-remaining": 1,
                "submit-time": "2023-05-25T20:05:05.564137Z",
                "finish-time": "None",
            }
        ]
    }


@fixture
def completed_status() -> TransformStatus:
    return TransformStatus(
        **{
            "request_id": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf",
            "did": "File List Provided in Request",
            "columns": None,
            "selection": "(Where (SelectMany (call EventDataset) (lambda (list e) (call (attr e 'Jets') 'AntiKt4EMTopoJets'))) (lambda (list j) (and (> (/ (call (attr j 'pt')) 1000) 20) (< (call abs (/ (call (attr j 'eta')) 1000)) 4.5))))",  # NOQA
            "tree-name": None,
            "image": "sslhep/servicex_func_adl_uproot_transformer:uproot4",
            "workers": None,
            "result-destination": "object-store",
            "result-format": "parquet",
            "workflow-name": "selection_codegen",
            "generated-code-cm": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf-generated-source",
            "status": "Submitted",
            "failure-info": None,
            "app-version": "develop",
            "code-gen-image": "sslhep/servicex_code_gen_func_adl_uproot:v1.2.0",
            "files": 1,
            "files-completed": 0,
            "files-failed": 0,
            "files-remaining": 1,
            "submit-time": "2023-05-25T20:05:05.564137Z",
            "finish-time": "None",
            "minio-endpoint": "minio.org:9000",
            "minio-secured": False,
            "minio-access-key": "miniouser",
            "minio-secret-key": "secret",
        }
    )


@fixture
def transformed_result(dummy_parquet_file) -> TransformedResults:
    return TransformedResults(
        hash="123-4455",
        title="Test",
        codegen="uproot",
        request_id="123-45-6789",
        submit_time=datetime.now(),
        data_dir="/foo/bar",
        file_list=[dummy_parquet_file],
        signed_url_list=[],
        files=2,
        result_format=ResultFormat.root,
    )


@fixture
def dummy_parquet_file():
    data = {'column1': [1, 2, 3, 4],
            'column2': ['A', 'B', 'C', 'D']}
    df = pd.DataFrame(data)
    parquet_file_path = '1.parquet'
    df.to_parquet(parquet_file_path, index=False)

    yield parquet_file_path

    if os.path.exists(parquet_file_path):
        os.remove(parquet_file_path)



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import abc
import asyncio
from abc import ABC
from asyncio import Task, CancelledError
from typing import List, Optional, Union
from servicex.expandable_progress import ExpandableProgress

try:
    import pandas as pd
except ModuleNotFoundError:
    pass


from servicex.types import DID

import rich
from rich.progress import (
    Progress,
    TaskID)

from servicex.configuration import Configuration
from servicex.minio_adapter import MinioAdapter
from servicex.models import (
    TransformRequest,
    ResultDestination,
    ResultFormat,
    Status,
    TransformedResults,
)
from servicex.query_cache import QueryCache
from servicex.servicex_adapter import ServiceXAdapter

from make_it_sync import make_sync

ProgressIndicators = Union[Progress, ExpandableProgress]


class Query(ABC):
    def __init__(
        self,
        dataset_identifier: DID,
        title: str,
        codegen: str,
        sx_adapter: ServiceXAdapter,
        config: Configuration,
        query_cache: QueryCache,
        servicex_polling_interval: int = 5,
        minio_polling_interval: int = 5,
        result_format: ResultFormat = ResultFormat.parquet,
        ignore_cache: bool = False,
    ):
        r"""
        This is the main class for constructing transform requests and receiving the
        results in a format of your choice. It can be run synchronously or asynchronously.

        :param dataset_identifier: Either a Rucio DID or a list of files
        :param title: Human readable title for this transform
        :param codegen: Name of the code generator
        :param sx_adapter:
        :param config:
        :param query_cache:
        :param servicex_polling_interval: How many seconds between polling for
                                          transform status?
        :param minio_polling_interval:  How many seconds between polling the minio bucket
                                        for new files?
        :param result_format:
        :param ignore_cache:  If true, ignore the cache and always submit a new transform
        """
        super(Query, self).__init__()
        self.servicex = sx_adapter
        self.configuration = config
        self.cache = query_cache

        self.dataset_identifier = dataset_identifier
        self.codegen = codegen
        self.title = title

        self.result_format = result_format
        self.signed_urls = False
        self.current_status = None
        self.download_path = None
        self.minio = None
        self.files_failed = None
        self.files_completed = None
        self._return_qastle = True

        self.request_id = None
        self.ignore_cache = ignore_cache

        # Number of seconds in between ServiceX status polls
        self.servicex_polling_interval = servicex_polling_interval
        self.minio_polling_interval = minio_polling_interval

    @abc.abstractmethod
    def generate_selection_string(self) -> str:
        pass

    @property
    def transform_request(self):
        if not self.result_format:
            raise ValueError(
                "Unable to determine the result file format. Use set_result_format method"
            )  # NOQA E501

        sx_request = TransformRequest(
            title=self.title,
            codegen=self.codegen,
            result_destination=ResultDestination.object_store,  # type: ignore
            result_format=self.result_format,  # type: ignore
            selection=self.generate_selection_string(),
        )  # type: ignore
        # Transfer the DID into the transform request
        self.dataset_identifier.populate_transform_request(sx_request)
        return sx_request

    def set_title(self, title: str) -> Query:
        self.title = title
        return self

    def set_result_format(self, result_format: ResultFormat):
        r"""
        Set the result format - required at constructor time or as part of the query
        chain of methods
        :param result_format:
        :return: self to allow you to chain together query and setup methods
        """
        self.result_format = result_format
        return self

    async def submit_and_download(
        self, signed_urls_only: bool,
            expandable_progress: ExpandableProgress,
            dataset_group: Optional[bool] = False,
    ) -> Optional[TransformedResults]:
        """
        Submit the transform request to ServiceX. Poll the transform status to see when
        the transform completes and to get the number of files in the dataset along with
        current progress and failed file count.

        :param signed_urls_only: Set to true to skip actually downloading the files and
                                 just return pre-signed urls
        :param display_progress: Set to false to disable the progress bar
        :param expandable_progress: Provide an existing progress bar. Set to None to have
                                    one created for you

        :return: Transform results object which contains the list of files downloaded
                 or the list of pre-signed urls
        """
        download_files_task = None
        loop = asyncio.get_running_loop()

        def transform_complete(task: Task):
            """
            Called when the Monitor task completes. This could be because of exception or
            the transform completed
            :param task:
            :return:
            """
            if task.exception():
                rich.print("ServiceX Exception", task.exception())
                if download_files_task:
                    download_files_task.cancel("Transform failed")
                raise task.exception()

            if self.current_status.files_failed:
                rich.print(
                    f"[bold red]Transforms completed with failures[/bold red] "
                    f"{self.current_status.files_failed} files failed out of "
                    f"{self.current_status.files}"
                )
            else:
                rich.print("Transforms completed successfully")

        sx_request = self.transform_request

        # Let's see if this is in the cache already, but respect the user's wishes
        # to ignore the cache
        cached_record = self.cache.get_transform_by_hash(sx_request.compute_hash()) \
            if not self.ignore_cache \
            else None

        # And that we grabbed the resulting files in the way that the user requested
        # (Downloaded, or obtained pre-signed URLs)
        if cached_record:
            if (
                signed_urls_only
                and cached_record.signed_url_list
                or not signed_urls_only
                and cached_record.file_list
            ):
                rich.print("Returning results from cache")
                return cached_record

        # If we get here with a cached record, then we know that the transform
        # has been run, but we just didn't get the files from object store in the way
        # requested by user
        transform_bar_title = "Transform"
        if not cached_record:
            transform_progress = expandable_progress.add_task(
                transform_bar_title, start=False, total=None
            ) if expandable_progress else None
        else:
            self.request_id = cached_record.request_id
            transform_progress = None
            await self.retrieve_current_transform_status()

        minio_progress_bar_title = (
            "Download" if not signed_urls_only else "Signing URLS"
        )

        download_progress = expandable_progress.add_task(
            minio_progress_bar_title, start=False, total=None
        ) if expandable_progress else None

        if not cached_record:
            self.request_id = await self.servicex.submit_transform(sx_request)

            monitor_task = loop.create_task(
                self.transform_status_listener(
                    expandable_progress, transform_progress, transform_bar_title,
                    download_progress, minio_progress_bar_title
                )
            )
            monitor_task.add_done_callback(transform_complete)
        else:
            self.request_id = cached_record.request_id

        download_files_task = loop.create_task(
            self.download_files(
                signed_urls_only, expandable_progress, download_progress, cached_record
            )
        )

        try:
            signed_urls = []
            downloaded_files = []
            download_result = await download_files_task
            if signed_urls_only:
                signed_urls = download_result
                if cached_record:
                    cached_record.signed_url_list = download_result
            else:
                downloaded_files = download_result
                if cached_record:
                    cached_record.file_list = download_result

            # Update the cache
            if not cached_record:
                transform_report = self.cache.cache_transform(
                    sx_request,
                    self.current_status,
                    self.download_path.as_posix(),
                    downloaded_files,
                    signed_urls,
                )
            else:
                self.cache.update_record(cached_record)
                transform_report = cached_record

            return transform_report
        except CancelledError:
            rich.print("Aborted file downloads due to transform failure")

    async def transform_status_listener(
        self, progress: ExpandableProgress, progress_task: TaskID,
        progress_bar_title: str, download_task: TaskID, download_bar_title: str
    ):
        """
        Poll ServiceX for the status of a transform. Update progress bars and keep track
        of status. Once we know the number of files in the dataset, update the progress
        bars.
        """

        # Actual number of files in the dataset. We only know this once the DID
        # finder has completed its work. In the meantime transformers will already
        # start up and begin work on the files we know about
        final_count = None

        while True:
            await self.retrieve_current_transform_status()

            # Do we finally know the final number of files in the dataset? Now is the
            # time to properly initialize the progress bars
            if not final_count and self.current_status.files:
                final_count = self.current_status.files
                if progress:
                    progress.update(progress_task, progress_bar_title, total=final_count)
                    progress.start_task(task_id=progress_task, task_type="Transform")

                    progress.update(download_task, download_bar_title, total=final_count)
                    progress.start_task(task_id=download_task, task_type="Download")

            if progress:
                progress.update(
                    progress_task, progress_bar_title,
                    completed=self.current_status.files_completed
                )

            if self.current_status.status == Status.complete:
                self.files_completed = self.current_status.files_completed
                self.files_failed = self.current_status.files_failed
                return

            await asyncio.sleep(self.servicex_polling_interval)

    async def retrieve_current_transform_status(self):
        s = await self.servicex.get_transform_status(self.request_id)

        # Is this the first time we've polled status? We now know the request ID.
        # Update the display and set our download directory.
        if not self.current_status:
            rich.print(f"[bold]ServiceX Transform {s.request_id}[/bold]")
            self.download_path = self.cache.cache_path_for_transform(s)

        self.current_status = s

        # We can only initialize the minio adapter with data from the transform
        # status. This includes the minio host and credentials. We use the
        # transform id as the bucket.
        if not self.minio:
            self.minio = MinioAdapter.for_transform(self.current_status)

    async def download_files(
        self,
        signed_urls_only: bool,
        progress: ExpandableProgress,
        download_progress: TaskID,
        cached_record: Optional[TransformedResults],
    ) -> List[str]:
        """
        Task to monitor the list of files in the transform output's bucket. Any new files
        will be downloaded.
        """
        files_seen = set()
        result_uris = []
        download_tasks = []
        loop = asyncio.get_running_loop()

        async def download_file(
            minio: MinioAdapter,
            filename: str,
            progress: Progress,
            download_progress: TaskID,
            shorten_filename: bool = False,
        ):
            downloaded_filename = await minio.download_file(
                filename, self.download_path, shorten_filename=shorten_filename
            )
            result_uris.append(downloaded_filename.as_posix())
            progress.advance(task_id=download_progress, task_type="Download")

        async def get_signed_url(
            minio: MinioAdapter,
            filename: str,
            progress: Optional[Progress],
            download_progress: TaskID,
        ):
            url = await minio.get_signed_url(filename)
            result_uris.append(url)
            if progress:
                progress.advance(task_id=download_progress, task_type="Download")

        while True:
            if not cached_record:
                await asyncio.sleep(self.minio_polling_interval)
            if self.minio:
                files = await self.minio.list_bucket()
                for file in files:
                    if file.filename not in files_seen:
                        if signed_urls_only:
                            download_tasks.append(
                                loop.create_task(
                                    get_signed_url(
                                        self.minio,
                                        file.filename,
                                        progress,
                                        download_progress,
                                    )
                                )
                            )
                        else:
                            download_tasks.append(
                                loop.create_task(
                                    download_file(
                                        self.minio,
                                        file.filename,
                                        progress,
                                        download_progress,
                                        shorten_filename=self.configuration.shortened_downloaded_filename,  # NOQA: E501
                                    )
                                )
                            )  # NOQA 501
                        files_seen.add(file.filename)

            # Once the transform is complete we can stop polling since all the files
            # are guaranteed to be in the bucket. Also, if we are just downloading or
            # signing urls for a previous transform then we know it is complete as well
            if cached_record or (
                self.current_status and self.current_status.status == Status.complete
            ):
                break

        # Now just wait until all of our tasks complete
        await asyncio.gather(*download_tasks)
        return result_uris

    async def as_files_async(self,
                             display_progress: bool = True,
                             provided_progress: Optional[ProgressIndicators] = None
                             ) -> TransformedResults:
        r"""
        Submit the transform and request all the resulting files to be downloaded
        :return: TransformResult instance with the list of complete paths to the downloaded files
        """
        with ExpandableProgress(display_progress, provided_progress) as progress:
            return await self.submit_and_download(signed_urls_only=False,
                                                  expandable_progress=progress)

    as_files = make_sync(as_files_async)

    try:
        async def as_pandas_async(self,
                                  display_progress: bool = True,
                                  provided_progress: Optional[ProgressIndicators] = None) \
                -> pd.DataFrame:
            r"""
            Return a pandas dataframe containing the results. This only works if you've
            installed pandas extra

            :return: Pandas Dataframe
            """
            self.result_format = ResultFormat.parquet
            transformed_result = await self.as_files_async(display_progress=display_progress,
                                                           provided_progress=provided_progress
                                                           )
            dataframes = pd.concat(
                [pd.read_parquet(p) for p in transformed_result.file_list]
            )
            return dataframes

        as_pandas = make_sync(as_pandas_async)
    except NameError:
        pass

    async def as_signed_urls_async(self, display_progress: bool = True,
                                   provided_progress: Optional[ProgressIndicators] = None,
                                   dataset_group: bool = False) \
            -> TransformedResults:
        r"""
        Presign URLs for each of the transformed files

        :return: TransformedResults object with the presigned_urls list populated
        """
        if dataset_group:
            return await self.submit_and_download(signed_urls_only=True,
                                                  expandable_progress=provided_progress,
                                                  dataset_group=dataset_group)

        with ExpandableProgress(display_progress=display_progress,
                                provided_progress=provided_progress) as progress:
            return await self.submit_and_download(signed_urls_only=True,
                                                  expandable_progress=progress,
                                                  dataset_group=dataset_group)

    as_signed_urls = make_sync(as_signed_urls_async)



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
from unittest.mock import patch

from servicex.configuration import Configuration


@patch('servicex.configuration.tempfile.gettempdir', return_value="./mytemp")
def test_config_read(tempdir):
    # Windows style user name
    os.environ['UserName'] = "p_higgs"
    c = Configuration.read(config_path="tests/example_config.yaml")
    assert c.cache_path == "mytemp/servicex_p_higgs"

    # Reset environment
    del os.environ['UserName']

    # Linux style user name
    os.environ['USER'] = "p_higgs2"
    c = Configuration.read(config_path="tests/example_config.yaml")
    assert c.cache_path == "mytemp/servicex_p_higgs2"


@patch('servicex.configuration.tempfile.gettempdir', return_value="./mytemp")
def test_default_cache_path(tempdir):

    # Windows style user name
    os.environ['UserName'] = "p_higgs"
    c = Configuration.read(config_path="tests/example_config_no_cache_path.yaml")
    assert c.cache_path == "mytemp/servicex_p_higgs"
    del os.environ['UserName']

    # Linux style user name
    os.environ['USER'] = "p_higgs"
    c = Configuration.read(config_path="tests/example_config_no_cache_path.yaml")
    assert c.cache_path == "mytemp/servicex_p_higgs"
    del os.environ['USER']



# Copyright (c) 2023, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# import pathlib
from typing import Any, Dict, List
from rich.progress import Progress, TextColumn, BarColumn, MofNCompleteColumn, \
    TimeRemainingColumn

from servicex.dataset_identifier import RucioDatasetIdentifier, FileListDataset
from servicex.servicex_client import ServiceXClient
from servicex.models import ResultFormat


class DataBinderRequests:
    """
    Prepare ServiceX requests from DataBinder configuration
    """

    def __init__(self, updated_config: Dict[str, Any]):
        self._config = updated_config
        self._client = self._get_client()
        self._progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TimeRemainingColumn(compact=True, elapsed_when_finished=True)
        )

    def _get_client(self):
        if 'http' in self._config['General']['ServiceX']:
            return ServiceXClient(backend=None, url=self._config['General']['ServiceX'])
        else:
            return ServiceXClient(backend=self._config['General']['ServiceX'], url=None)

    def get_requests(self) -> List:
        list_requests = []
        for sample in self._config['Sample']:
            list_requests.append(self._build_request(sample))
        # flatten nested lists for samples with more than two Rucio datasets
        flist_requests = [request for x in list_requests for request in x]
        return flist_requests

    def _build_request(self, sample: Dict[str, Any]):
        """
        Return a list containing ServiceX request(s) of the given sample
        """
        requests_sample = []

        def _get_input_source(sample):
            if 'RucioDID' in sample.keys():
                nfiles = None
                if 'NFiles' in sample.keys():
                    nfiles = int(sample['NFiles'])
                input_source = RucioDatasetIdentifier(
                    str(sample['RucioDID']),
                    num_files=nfiles or 0
                )
            elif 'XRootDFiles' in sample.keys():
                input_source = FileListDataset(
                    [file.strip() for file in sample['XRootDFiles'].split(",")]
                )
            else:
                raise TypeError(
                    f"Unknown input source in Sample {sample['Name']}"
                )
            return input_source

        def _set_result_format():
            if self._config['General']['OutputFormat'].lower() == "root":
                return ResultFormat.root
            elif self._config['General']['OutputFormat'].lower() == "parquet":
                return ResultFormat.parquet
            else:
                raise ValueError(
                    f"Output format {self._config['General']['OutputFormat']} is not supported"
                )

        def _get_servicex_dataset(sample):
            if sample['Codegen'] == 'python':
                return self._client.python_dataset(
                    dataset_identifier=_get_input_source(sample),
                    title=sample['Name'],
                    codegen="python",
                    ignore_cache=sample['IgnoreLocalCache'],
                    result_format=_set_result_format()
                )
            else:
                raise TypeError(
                    f"Unknown code-generator in Sample {sample['Name']}"
                )

        def _servicex_dataset_query(sample):
            if sample['Codegen'] == 'python':
                query = sample['Function']
                return _get_servicex_dataset(sample).with_uproot_function(
                    query
                )
            else:
                raise TypeError(
                    f"Unknown code-generator in Sample {sample['Name']}"
                )

        requests_sample.append(
            {
                "sample_name": sample['Name'],
                "delivery": self._config['General']['Delivery'].lower(),
                "ds_query": _servicex_dataset_query(sample)
            }
        )
        return requests_sample



import pytest
from pydantic import ValidationError

from servicex import ServiceXSpec, FileListDataset, RucioDatasetIdentifier


def basic_spec(samples=None):
    return {
        "General": {
            "ServiceX": "servicex",
            "Codegen": "python",
        },
        "Sample": samples
        or [{"Name": "sampleA", "XRootDFiles": "root://a.root", "Function": "a"}],
    }


def test_load_config():
    config = {
        "General": {
            "ServiceX": "servicex",
            "Codegen": "python",
            "Delivery": "LocalCache",
        },
        "Sample": [
            {"Name": "sampleA", "RucioDID": "user.kchoi:sampleA", "Function": "DEF_a"},
            {
                "Name": "sampleB",
                "XRootDFiles": "root://a.root",
                "Columns": "el_pt",
                "Codegen": "uproot",
                "Function": "DEF_a",
            },
        ],
        "Definition": {"DEF_a": "a"},
    }
    new_config = ServiceXSpec.parse_obj(config)
    assert new_config.Sample[0].Function == "a"


def test_single_root_file():

    spec = ServiceXSpec.parse_obj(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "XRootDFiles": "root://eospublic.cern.ch//file1.root",
                    "Function": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, FileListDataset)
    assert spec.Sample[0].dataset_identifier.files == [
        "root://eospublic.cern.ch//file1.root"
    ]


def test_list_of_root_files():
    spec = ServiceXSpec.parse_obj(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "XRootDFiles": [
                        "root://eospublic.cern.ch//file1.root",
                        "root://eospublic.cern.ch//file2.root",
                    ],
                    "Function": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, FileListDataset)
    assert spec.Sample[0].dataset_identifier.files == [
        "root://eospublic.cern.ch//file1.root",
        "root://eospublic.cern.ch//file2.root",
    ]


def test_rucio_did():
    spec = ServiceXSpec.parse_obj(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "Function": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, RucioDatasetIdentifier)
    assert (
        spec.Sample[0].dataset_identifier.did
        == "rucio://user.ivukotic:user.ivukotic.single_top_tW__nominal"
    )


def test_rucio_did_numfiles():
    spec = ServiceXSpec.parse_obj(
        basic_spec(
            samples=[
                {
                    "Name": "sampleA",
                    "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                    "NFiles": 10,
                    "Function": "a",
                }
            ]
        )
    )

    assert isinstance(spec.Sample[0].dataset_identifier, RucioDatasetIdentifier)
    assert (
        spec.Sample[0].dataset_identifier.did
        == "rucio://user.ivukotic:user.ivukotic.single_top_tW__nominal?files=10"
    )


def test_invalid_dataset_identifier():
    with pytest.raises(ValidationError):
        ServiceXSpec.parse_obj(
            basic_spec(
                samples=[
                    {
                        "Name": "sampleA",
                        "RucioDID": "user.ivukotic:user.ivukotic.single_top_tW__nominal",
                        "XRootDFiles": "root://eospublic.cern.ch//file1.root",
                        "NFiles": 10,
                        "Function": "a",
                    }
                ]
            )
        )

    with pytest.raises(ValidationError):
        ServiceXSpec.parse_obj(
            basic_spec(
                samples=[
                    {
                        "Name": "sampleA",
                        "NFiles": 10,
                        "Function": "a",
                    }
                ]
            )
        )



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import shutil

import rich
import typer
from rich.prompt import Confirm
from rich.table import Table

from servicex.servicex_client import ServiceXClient

cache_app = typer.Typer(name="cache", no_args_is_help=True)


@cache_app.callback()
def cache():
    """
    Sub-commands for creating and manipulating the local query cache
    """
    pass


@cache_app.command()
def list():
    """
    List the cached queries
    """
    sx = ServiceXClient()
    cache = sx.query_cache
    table = Table(title="Cached Queries")
    table.add_column("Title")
    table.add_column("Codegen")
    table.add_column("Transform ID")
    table.add_column("Run Date")
    table.add_column("Files")
    table.add_column("Format")
    runs = cache.cached_queries()
    for r in runs:
        table.add_row(
            r.title,
            r.codegen,
            r.request_id,
            r.submit_time.astimezone().strftime("%a, %Y-%m-%d %H:%M"),
            str(r.files),
            r.result_format
        )
    rich.print(table)


@cache_app.command()
def clear(force: bool = typer.Option(False, "-y", help="Force, don't ask for permission")):
    """
    Clear the local query cache
    """
    if force or Confirm.ask("Really clear cache and delete downloaded files?"):
        sx = ServiceXClient()
        sx.query_cache.close()
        shutil.rmtree(sx.config.cache_path)
        rich.print("Cache cleared")


@cache_app.command(no_args_is_help=True)
def delete(transform_id: str = typer.Option(None, "-t", "--transform-id", help="Transform ID")):
    """
    Delete a cached query. Use -t to specify the transform ID
    """
    sx = ServiceXClient()
    cache = sx.query_cache
    rec = cache.get_transform_by_request_id(transform_id)
    shutil.rmtree(rec.data_dir)
    cache.delete_record_by_request_id(rec.request_id)



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional, List, TypeVar, Any, Type

from servicex.configuration import Configuration
from servicex.func_adl.func_adl_dataset import FuncADLQuery
from servicex.models import ResultFormat, TransformStatus
from servicex.query_cache import QueryCache
from servicex.servicex_adapter import ServiceXAdapter
from servicex.types import DID
from servicex.python_dataset import PythonQuery
from servicex.dataset_group import DatasetGroup
import ast
import qastle

from make_it_sync import make_sync
from servicex.databinder_models import ServiceXSpec, General, Sample

import rich

T = TypeVar("T")


def deliver(config: ServiceXSpec):
    def get_codegen(_sample: Sample, _general: General):
        if _sample.Codegen:
            return _sample.Codegen
        else:
            return _general.Codegen

    sx = ServiceXClient(backend=config.General.ServiceX)
    datasets = []
    for sample in config.Sample:
        if sample.Query:
            if type(sample.Query) is str:
                qastle_query = qastle.python_ast_to_text_ast(ast.parse(sample.Query)) # NOQA E501
                sample.Query = FuncADLQuery()

                sample.Query.set_provided_qastle(qastle_query)

            query = sx.func_adl_dataset(sample.dataset_identifier, sample.Name,
                                        get_codegen(sample, config.General),
                                        config.General.OutputFormat)
            query._q_ast = sample.Query._q_ast
            query._item_type = sample.Query._item_type
            if sample.Tree:
                query = query.set_tree(sample.Tree)

            sample.Query = query
            sample.Query.ignore_cache = sample.IgnoreLocalCache

            datasets.append(sample.Query)
        elif sample.Function:
            # The function field can be a callable if this was all initialized
            # in python, or it can be a string if it was initialized from a
            # yaml file. If it comes from a string, let's validate the syntax
            # now to avoid nasty surprises later.
            if isinstance(sample.Function, str):
                try:
                    exec(sample.Function)
                except SyntaxError as e:
                    raise SyntaxError(f"Syntax error in {sample.Name}: {e}")

            dataset = sx.python_dataset(sample.dataset_identifier, sample.Name,
                                        get_codegen(sample, config.General),
                                        config.General.OutputFormat)
            dataset.python_function = sample.Function
            dataset.ignore_cache = sample.IgnoreLocalCache
            datasets.append(dataset)

    group = DatasetGroup(datasets)

    if config.General.Delivery == General.DeliveryEnum.SignedURLs:
        results = group.as_signed_urls()
        return {obj.title: obj.signed_url_list for obj in results}

    elif config.General.Delivery == General.DeliveryEnum.LocalCache:
        results = group.as_files()
        return {obj.title: obj.file_list for obj in results}


class ServiceXClient:
    r"""
    Connection to a ServiceX deployment. Instances of this class can deployment
    data from the service and also interact with previously run transformations.
    Instances of this class are factories for `Datasets``
    """

    def __init__(self, backend=None, url=None, config_path=None):
        r"""
        If both `backend` and `url` are unspecified then it will attempt to pick up
        the default backend from `.servicex`

        :param backend: Name of a deployment from the .servicex file
        :param url: Direct URL of a serviceX deployment instead of using .servicex.
                    Can only work with hosts without auth, or the token is found
                    in a file pointed to by the environment variable BEARER_TOKEN_FILE
        :param config_path: Optional path te the `.servicex` file. If not specified,
                    will search in local directory and up in enclosing directories
        """
        self.config = Configuration.read(config_path)
        self.endpoints = self.config.endpoint_dict()

        if not url and not backend:
            backend = self.config.default_endpoint

        if bool(url) == bool(backend):
            raise ValueError("Only specify backend or url... not both")

        if url:
            self.servicex = ServiceXAdapter(url)
        elif backend:
            if backend not in self.endpoints:
                raise ValueError(f"Backend {backend} not defined in .servicex file")
            self.servicex = ServiceXAdapter(
                self.endpoints[backend].endpoint,
                refresh_token=self.endpoints[backend].token,
            )

        self.query_cache = QueryCache(self.config)
        self.code_generators = set(self.get_code_generators(backend).keys())

    async def get_transforms_async(self) -> List[TransformStatus]:
        r"""
        Retrieve all transforms you have run on the server
        :return: List of Transform status objects
        """
        return await self.servicex.get_transforms()

    get_transforms = make_sync(get_transforms_async)

    async def get_transform_status_async(self, transform_id) -> TransformStatus:
        r"""
        Get the status of a given transform
        :param transform_id: The uuid of the transform
        :return: The current status for the transform
        """
        return await self.servicex.get_transform_status(request_id=transform_id)

    get_transform_status = make_sync(get_transform_status_async)

    def get_code_generators(self, backend=None):
        r"""
        Retrieve the code generators deployed with the serviceX instance
        :return:  The list of code generators as json dictionary
        """
        cached_backends = None
        if backend:
            cached_backends = self.query_cache.get_codegen_by_backend(backend)
        if cached_backends:
            rich.print("Returning code generators from cache")
            return cached_backends["codegens"]
        else:
            code_generators = self.servicex.get_code_generators()
            self.query_cache.update_codegen_by_backend(backend, code_generators)
            return code_generators

    def func_adl_dataset(
        self,
        dataset_identifier: DID,
        title: str = "ServiceX Client",
        codegen: str = "uproot",
        result_format: Optional[ResultFormat] = None,
        item_type: Type[T] = Any,
        ignore_cache: bool = False
    ) -> FuncADLQuery[T]:
        r"""
        Generate a dataset that can use func_adl query language

        :param dataset_identifier:  The dataset identifier or filelist to be the source of files
        :param title: Title to be applied to the transform. This is also useful for
                      relating transform results.
        :param codegen: Name of the code generator to use with this transform
        :param result_format:  Do you want Paqrquet or Root? This can be set later with
                               the set_result_format method
        :param item_type: The type of the items that will be returned from the query
        :param ignore_cache: Ignore the query cache and always run the query
        :return: A func_adl dataset ready to accept query statements.
        """
        if codegen not in self.code_generators:
            raise NameError(
                f"{codegen} code generator not supported by serviceX "
                f"deployment at {self.servicex.url}"
            )

        return FuncADLQuery(
            dataset_identifier,
            sx_adapter=self.servicex,
            title=title,
            codegen=codegen,
            config=self.config,
            query_cache=self.query_cache,
            result_format=result_format,
            item_type=item_type,
            ignore_cache=ignore_cache
        )

    def python_dataset(
        self,
        dataset_identifier: DID,
        title: str = "ServiceX Client",
        codegen: str = "uproot",
        result_format: Optional[ResultFormat] = None,
        ignore_cache: bool = False
    ) -> PythonQuery:
        r"""
        Generate a dataset that can use accept a python function for the  query

        :param dataset_identifier:  The dataset identifier or filelist to be the source of files
        :param title: Title to be applied to the transform. This is also useful for
                      relating transform results.
        :param codegen: Name of the code generator to use with this transform
        :param result_format:  Do you want Paqrquet or Root? This can be set later with
                               the set_result_format method
        :param ignore_cache: Ignore the query cache and always run the query
        :return: A func_adl dataset ready to accept a python function statements.

        """

        if codegen not in self.code_generators:
            raise NameError(
                f"{codegen} code generator not supported by serviceX "
                f"deployment at {self.servicex.url}"
            )

        return PythonQuery(
            dataset_identifier,
            sx_adapter=self.servicex,
            title=title,
            codegen=codegen,
            config=self.config,
            query_cache=self.query_cache,
            result_format=result_format,
            ignore_cache=ignore_cache
        )



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import tempfile

import pytest

from servicex.configuration import Configuration
from servicex.models import ResultFormat
from servicex.query_cache import QueryCache, CacheException
from tinydb import Query

file_uris = ["/tmp/foo1.root", "/tmp/foo2.root"]


def test_hash(transform_request):
    request1 = transform_request.copy()
    request2 = transform_request.copy()

    assert request1.compute_hash() == request2.compute_hash()

    # Changing the title doesn't impact the hash
    request2.title = "This has changed"
    assert request1.compute_hash() == request2.compute_hash()

    # Changing the DID does
    request2.did = "rucio://baz.bar"
    assert request1.compute_hash() != request2.compute_hash()

    # Changing the did backend does
    request2 = request1.copy()
    request2.codegen = "foo"
    assert request1.compute_hash() != request2.compute_hash()

    # Changing result_format does
    request2 = request1.copy()
    request2.result_format = ResultFormat.root
    assert request1.compute_hash() != request2.compute_hash()


def test_cache_transform(transform_request, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        cache.cache_transform(
            transform=transform_request,
            completed_status=completed_status,
            data_dir="/foo/bar",
            file_list=file_uris,
            signed_urls=[],
        )

        test = cache.get_transform_by_hash(transform_request.compute_hash())
        assert test
        assert test.title == "Test submission"
        assert test.request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"

        test2 = cache.get_transform_by_request_id(
            "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"
        )
        assert test2
        assert test2.title == "Test submission"
        assert test2.request_id == "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"

        assert not cache.get_transform_by_hash("thishashdoesnotexist")
        assert not cache.get_transform_by_request_id("this-uuid-does-not-exist")

        # make a duplicate record
        cache.cache_transform(
            transform=transform_request,
            completed_status=completed_status,
            data_dir="/foo/baz",
            file_list=file_uris,
            signed_urls=[],
        )

        with pytest.raises(CacheException):
            _ = cache.get_transform_by_hash(transform_request.compute_hash())

        with pytest.raises(CacheException):
            _ = cache.get_transform_by_request_id(
                "b8c508d0-ccf2-4deb-a1f7-65c839eebabf"
            )

        assert len(cache.cached_queries()) == 2

        cache.delete_record_by_request_id("b8c508d0-ccf2-4deb-a1f7-65c839eebabf")
        assert len(cache.cached_queries()) == 0

        cache.close()


def test_cache_path(completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        path_bits = os.path.split(cache.cache_path_for_transform(completed_status))
        assert path_bits[0] == temp_dir
        assert path_bits[1] == completed_status.request_id
        cache.close()


def test_record_delete(transform_request, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        cache.cache_transform(
            transform=transform_request,
            completed_status=completed_status,
            data_dir="/foo/bar",
            file_list=file_uris,
            signed_urls=[],
        )

        cache.cache_transform(
            transform=transform_request,
            completed_status=completed_status.copy(
                update={"request_id": "02c64494-4529-49a7-a4a6-95661ea3936e"}
            ),
            data_dir="/foo/baz",
            file_list=file_uris,
            signed_urls=[],
        )

        assert len(cache.cached_queries()) == 2
        cache.delete_record_by_request_id("02c64494-4529-49a7-a4a6-95661ea3936e")
        assert len(cache.cached_queries()) == 1
        cache.close()


def test_get_codegen_by_backend_empty():
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        result = cache.get_codegen_by_backend("non-existent")
        assert result is None
        cache.close()


def test_update_codegen_by_backend_single():
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        codegens = Query()
        cache.update_codegen_by_backend('backend_1', ['codegen_1'])
        result = cache.db.search(codegens.backend == 'backend_1')
        assert len(result) == 1
        assert result[0] == {'backend': 'backend_1', 'codegens': ['codegen_1']}
        cache.close()


def test_get_codegen_by_backend_single():
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        cache.update_codegen_by_backend('backend_1', ['codegen_1'])
        result = cache.get_codegen_by_backend("backend_1")
        assert result == {'backend': 'backend_1', 'codegens': ['codegen_1']}
        cache.close()


def test_delete_codegen_by_backend():
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        cache.update_codegen_by_backend('backend_1', ['codegen_1'])
        result = cache.get_codegen_by_backend("backend_1")
        assert result == {'backend': 'backend_1', 'codegens': ['codegen_1']}

        cache.delete_codegen_by_backend('backend_1')
        result = cache.get_codegen_by_backend("backend_1")
        assert result is None
        cache.close()


def test_delete_codegen_by_backend_nonexistent():
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        cache.delete_codegen_by_backend('backend_1')
        with pytest.raises(Exception):
            assert False
        cache.close()


def test_add_both_codegen_and_transform_to_cache(transform_request, completed_status):
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])  # type: ignore
        cache = QueryCache(config)
        cache.cache_transform(
            transform=transform_request,
            completed_status=completed_status,
            data_dir="/foo/bar",
            file_list=file_uris,
            signed_urls=[],
        )

        cache.update_codegen_by_backend('backend_1', ['codegen_1'])
        result = cache.get_codegen_by_backend("backend_1")
        assert result == {'backend': 'backend_1', 'codegens': ['codegen_1']}
        assert len(cache.cached_queries()) == 1
        cache.close()



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
from datetime import datetime
from typing import Optional, Dict, List

import httpx
from google.auth import jwt

from servicex.models import TransformRequest, TransformStatus


class AuthorizationError(BaseException):
    pass


class ServiceXAdapter:
    def __init__(self, url: str, refresh_token: Optional[str] = None):
        self.url = url
        self.refresh_token = refresh_token
        self.token = None

    async def _get_token(self, client: httpx.AsyncClient):
        url = f"{self.url}/token/refresh"
        headers = {"Authorization": f"Bearer {self.refresh_token}"}
        r = await client.post(url, headers=headers, json=None)
        if r.status_code == 200:
            self.token = r.json()['access_token']
        else:
            raise AuthorizationError(
                f"ServiceX access token request rejected: {r.status_code}"
            )

    @staticmethod
    def _get_bearer_token_file():
        bearer_token_file = os.environ.get('BEARER_TOKEN_FILE')
        bearer_token = None
        if bearer_token_file:
            with open(bearer_token_file, "r") as f:
                bearer_token = f.read().strip()
        return bearer_token

    async def _get_authorization(self, client: httpx.AsyncClient) -> Dict[str, str]:
        bearer_token = self._get_bearer_token_file()

        if bearer_token:
            self.token = bearer_token
        if not bearer_token and not self.refresh_token:
            return {}

        now = datetime.utcnow().timestamp()
        if not self.token or \
                float(jwt.decode(self.token, verify=False)["exp"]) - now < 0:
            await self._get_token(client)
        return {"Authorization": f"Bearer {self.token}"}

    async def get_transforms(self) -> List[TransformStatus]:
        async with httpx.AsyncClient() as client:
            headers = await self._get_authorization(client)
            r = await client.get(url=f"{self.url}/servicex/transformation",
                                 headers=headers)
            if r.status_code == 401:
                raise AuthorizationError(f"Not authorized to access serviceX at {self.url}")

            statuses = [TransformStatus(**status) for status in r.json()['requests']]
        return statuses

    def get_code_generators(self):
        with httpx.Client() as client:
            r = client.get(url=f"{self.url}/multiple-codegen-list")

            if r.status_code == 403:
                raise AuthorizationError(
                    f"Not authorized to access serviceX at {self.url}")
        return r.json()

    async def submit_transform(self, transform_request: TransformRequest):
        async with httpx.AsyncClient() as client:
            headers = await self._get_authorization(client)
            r = await client.post(url=f"{self.url}/servicex/transformation",
                                  headers=headers,
                                  json=transform_request.dict(by_alias=True,
                                                              exclude_none=True))
            if r.status_code == 401:
                raise AuthorizationError(
                    f"Not authorized to access serviceX at {self.url}")
            elif r.status_code == 400:
                raise ValueError(f"Invalid transform request: {r.json()['message']}")
        return r.json()['request_id']

    async def get_transform_status(self, request_id: str) -> TransformStatus:
        async with httpx.AsyncClient() as client:
            headers = await self._get_authorization(client)
            r = await client.get(url=f"{self.url}/servicex/transformation/{request_id}",
                                 headers=headers)
            if r.status_code == 401:
                raise AuthorizationError(f"Not authorized to access serviceX at {self.url}")
            if r.status_code == 404:
                raise ValueError(f"Transform ID {request_id} not found")
            status = TransformStatus(**r.json())
            return status



# Copyright (c) 2023, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import ast
import copy
from typing import (
    Optional,
    Any,
    Type,
    TypeVar,
    cast,
    List,
    Union,
    Callable,
    Iterable,
    Dict,
)
from qastle import python_ast_to_text_ast

from func_adl import EventDataset, find_EventDataset
from func_adl.object_stream import S
from servicex.configuration import Configuration
from servicex.query import Query
from servicex.func_adl.util import has_tuple
from servicex.models import ResultFormat
from servicex.query_cache import QueryCache
from servicex.servicex_adapter import ServiceXAdapter
from servicex.types import DID

T = TypeVar("T")


class FuncADLQuery(Query, EventDataset[T]):
    r"""
    ServiceX Dataset class that uses func_adl query syntax.
    """
    # These are methods that are translated locally
    _execute_locally = ["ResultPandasDF", "ResultAwkwardArray"]

    async def execute_result_async(
        self, a: ast.AST, title: Optional[str] = None
    ) -> Any:
        pass

    def check_data_format_request(self, f_name: str):
        pass

    def __init__(
        self,
        dataset_identifier: DID = None,
        sx_adapter: ServiceXAdapter = None,
        title: str = "ServiceX Client",
        codegen: str = None,
        config: Configuration = None,
        query_cache: QueryCache = None,
        result_format: Optional[ResultFormat] = None,
        item_type: Type = Any,
        ignore_cache: bool = False,
    ):
        Query.__init__(
            self,
            dataset_identifier=dataset_identifier,
            title=title,
            codegen=codegen,
            sx_adapter=sx_adapter,
            config=config,
            query_cache=query_cache,
            result_format=result_format,
            ignore_cache=ignore_cache,
        )
        EventDataset.__init__(self, item_type=item_type)
        self.provided_qastle = None

    def set_provided_qastle(self, qastle: str):
        self.provided_qastle = qastle

    def __deepcopy__(self, memo):
        """
        Customize deepcopy behavior for this class.
        We need to be careful because the query cache is a tinyDB database that holds an
        open file pointer. We are not allowed to clone an open file handle, so for this
        property we will copy by reference and share it between the clones
        """
        cls = self.__class__
        obj = cls.__new__(cls)

        memo[id(self)] = obj

        for attr, value in vars(self).items():
            if type(value) is QueryCache:
                setattr(obj, attr, value)
            else:
                setattr(obj, attr, copy.deepcopy(value, memo))

        return obj

    def SelectMany(
        self, func: Union[str, ast.Lambda, Callable[[T], Iterable[S]]]
    ) -> FuncADLQuery[S]:
        r"""
        Given the current stream's object type is an array or other iterable, return
        the items in this objects type, one-by-one. This has the effect of flattening a
        nested array.

        Arguments:

            func:   The function that should be applied to this stream's objects to return
                    an iterable. Each item of the iterable is now the stream of objects.

        Returns:
            The Dataset with the SelectMany operator applied

        Notes:
            - The function can be a `lambda`, the name of a one-line function, a string that
              contains a lambda definition, or a python `ast` of type `ast.Lambda`.
        """
        return super().SelectMany(func)

    def Select(self, f: Union[str, ast.Lambda, Callable[[T], S]]) -> FuncADLQuery[S]:
        r"""
        Apply a transformation function to each object in the stream, yielding a new type of
        object. There is a one-to-one correspondence between the input objects and output objects.

        Arguments:
            f:      selection function (lambda)

        Returns:
            The Dataset with the Select operator applied

        Notes:
            - The function can be a `lambda`, the name of a one-line function, a string that
              contains a lambda definition, or a python `ast` of type `ast.Lambda`.
        """

        return super().Select(f)

    def generate_selection_string(self) -> str:
        if self.provided_qastle:
            return self.provided_qastle
        else:
            return self.generate_qastle(self.query_ast)

    def Where(
        self, filter: Union[str, ast.Lambda, Callable[[T], bool]]
    ) -> FuncADLQuery[T]:
        r"""
        Filter the object stream, allowing only items for which `filter` evaluates as true through.

        Arguments:

            filter      A filter lambda that returns True/False.

        Returns:

            The Dataset with the Where operator applied

        Notes:
            - The function can be a `lambda`, the name of a one-line function, a string that
              contains a lambda definition, or a python `ast` of type `ast.Lambda`.
        """

        return super().Where(filter)

    def MetaData(self, metadata: Dict[str, Any]) -> FuncADLQuery[T]:
        r"""Add metadata to the current object stream. The metadata is an arbitrary set of string
        key-value pairs. The backend must be able to properly interpret the metadata.

        Returns:
            The Dataset with the MetaData operator applied
        """

        return super().MetaData(metadata)

    def QMetaData(self, metadata: Dict[str, Any]) -> FuncADLQuery[T]:
        r"""Add query metadata to the current object stream.

        - Metadata is never transmitted to any back end
        - Metadata is per-query, not per sample.

        Warnings are issued if metadata is overwriting metadata.

        Args:
            metadata (Dict[str, Any]): Metadata to be used later

        Returns:
            The Dataset with the QMetaData operator applied
        """
        return super().QMetaData(metadata)

    def set_tree(self, tree_name: str) -> FuncADLQuery[T]:
        r"""Set the tree name for the query.
        Args:
            tree_name (str): Name of the tree to use for the query

        Returns:
            The Dataset with the tree appended to the first call object
        """

        clone = self.clone_with_new_ast(copy.deepcopy(self.query_ast), self._item_type)
        c = find_EventDataset(clone.query_ast)
        c.args.append(ast.Str(s="bogus.root"))
        c.args.append(ast.Str(s=tree_name))
        return clone

    def generate_qastle(self, a: ast.AST) -> str:
        r"""Generate the qastle from the ast of the query.

        1. The top level function is already marked as being "ok"
        1. If the top level function is something we have to process locally,
           then we strip it off.

        Args:
            a (ast.AST): The complete AST of the request.

        Returns:
            str: Qastle that should be sent to servicex
        """
        top_function = cast(ast.Name, a.func).id
        source = a
        if top_function in self._execute_locally:
            # Request the default type here
            default_format = self._ds.first_supported_datatype(["parquet", "root-file"])
            assert default_format is not None, "Unsupported ServiceX returned format"
            method_to_call = self._format_map[default_format]

            stream = a.args[0]
            col_names = a.args[1]
            if method_to_call == "get_data_rootfiles_async":
                # If we have no column names, then we must be using a dictionary to
                # set them - so just pass that
                # directly.
                assert isinstance(
                    col_names, (ast.List, ast.Constant, ast.Str)
                ), f"Programming error - type name not known {type(col_names).__name__}"
                if isinstance(col_names, ast.List) and len(col_names.elts) == 0:
                    source = stream
                else:
                    source = ast.Call(
                        func=ast.Name(id="ResultTTree", ctx=ast.Load()),
                        args=[
                            stream,
                            col_names,
                            ast.Str("treeme"),
                            ast.Str("junk.root"),
                        ],
                    )
            elif method_to_call == "get_data_parquet_async":
                source = stream
                # See #32 for why this is commented out
                # source = ast.Call(
                #     func=ast.Name(id='ResultParquet', ctx=ast.Load()),
                #     args=[stream, col_names, ast.Str('junk.parquet')])
            else:  # pragma: no cover
                # This indicates a programming error
                assert False, f"Do not know how to call {method_to_call}"

        elif top_function == "ResultParquet":
            # Strip off the Parquet function, do a select if there are arguments for column names
            source = a.args[0]
            col_names = cast(ast.List, a.args[1]).elts

            def encode_as_tuple_reference(c_names: List) -> List[ast.AST]:
                # Encode each column ref as a index into the tuple we are being passed
                return [
                    ast.Subscript(
                        value=ast.Name(id="x", ctx=ast.Load()),
                        slice=ast.Constant(idx),
                        ctx=ast.Load(),
                    )
                    for idx, _ in enumerate(c_names)
                ]

            def encode_as_single_reference():
                # Single reference for a bare (non-col) variable
                return [
                    ast.Name(id="x", ctx=ast.Load()),
                ]

            if len(col_names) > 0:
                # Add a select on top to set the column names
                if len(col_names) == 1:
                    # Determine if they built a tuple or not
                    values = (
                        encode_as_tuple_reference(col_names)
                        if has_tuple(source)
                        else encode_as_single_reference()
                    )
                elif len(col_names) > 1:
                    values = encode_as_tuple_reference(col_names)
                else:
                    assert False, "make sure that type checkers can figure this out"

                d = ast.Dict(keys=col_names, values=values)
                tup_func = ast.Lambda(
                    args=ast.arguments(args=[ast.arg(arg="x")]), body=d
                )
                c = ast.Call(
                    func=ast.Name(id="Select", ctx=ast.Load()),
                    args=[source, tup_func],
                    keywords=[],
                )
                source = c

        return python_ast_to_text_ast(source)

    def as_qastle(self):
        r"""
        Generate Qastle from this AST

        :returns:
            Qastle representation of the target's AST
        """
        return self.value()



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import tempfile
from typing import Any
from unittest.mock import AsyncMock

import pytest

from servicex.configuration import Configuration
from servicex.dataset_identifier import FileListDataset
from servicex.expandable_progress import ExpandableProgress
from servicex.func_adl.func_adl_dataset import FuncADLQuery
from servicex.models import TransformStatus, Status, ResultFile, ResultFormat
from servicex.query_cache import QueryCache

transform_status = TransformStatus(
    **{
        "request_id": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf",
        "did": "File List Provided in Request",
        "columns": None,
        "selection": "(Where (SelectMany (call EventDataset) (lambda (list e) (call (attr e 'Jets') 'AntiKt4EMTopoJets'))) (lambda (list j) (and (> (/ (call (attr j 'pt')) 1000) 20) (< (call abs (/ (call (attr j 'eta')) 1000)) 4.5))))",  # NOQA 501
        "tree-name": None,
        "image": "sslhep/servicex_func_adl_uproot_transformer:uproot4",
        "workers": None,
        "result-destination": "object-store",
        "result-format": "parquet",
        "workflow-name": "selection_codegen",
        "generated-code-cm": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf-generated-source",
        "status": "Complete",
        "failure-info": None,
        "app-version": "develop",
        "code-gen-image": "sslhep/servicex_code_gen_func_adl_uproot:v1.2.0",
        "files": 0,
        "files-completed": 0,
        "files-failed": 0,
        "files-remaining": None,
        "submit-time": "2023-05-25T20:05:05.564137Z",
        "finish-time": "None",
        "minio-endpoint": "minio.org:9000",
        "minio-secured": False,
        "minio-access-key": "miniouser",
        "minio-secret-key": "letmein",
    }
)

transform_status1 = transform_status.copy(
    update={
        "status": Status.running,
        "files-remaining": None,
        "files-completed": 0,
        "files": 0,
    }
)
transform_status2 = transform_status.copy(
    update={
        "status": Status.running,
        "files-remaining": 1,
        "files-completed": 1,
        "files": 2,
    }
)
transform_status3 = transform_status.copy(
    update={
        "status": Status.complete,
        "files-remaining": 0,
        "files-completed": 2,
        "files": 2,
    }
)

file1 = ResultFile(filename="file1", size=100, extension="parquet")
file2 = ResultFile(filename="file2", size=100, extension="parquet")


@pytest.mark.asyncio
async def test_submit(mocker):
    servicex = AsyncMock()
    servicex.submit_transform = AsyncMock()
    servicex.submit_transform.return_value = {"request_id": '123-456-789"'}
    servicex.get_transform_status = AsyncMock()
    servicex.get_transform_status.side_effect = [
        transform_status1,
        transform_status2,
        transform_status3,
    ]

    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(side_effect=[[file1], [file1, file2]])
    mock_minio.download_file = AsyncMock()

    mock_cache = mocker.MagicMock(QueryCache)
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")
    datasource = FuncADLQuery(
        dataset_identifier=did,
        codegen="uproot",
        sx_adapter=servicex,
        query_cache=mock_cache,
    )
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        _ = await datasource.submit_and_download(signed_urls_only=False,
                                                 expandable_progress=progress)
        print(mock_minio.download_file.call_args)


def test_transform_request():
    servicex = AsyncMock()
    did = FileListDataset("/foo/bar/baz.root")
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        datasource = FuncADLQuery(
            dataset_identifier=did,
            codegen="uproot",
            sx_adapter=servicex,
            config=config,
            query_cache=cache,
        )

        q = (
            datasource.Select(lambda e: {"lep_pt": e["lep_pt"]})
            .set_result_format(ResultFormat.parquet)
            .transform_request
        )
        print("Qastle is ", q)
        cache.close()


def test_type():
    "Test that the item type for a dataset is correctly propagated"

    class my_type_info:
        "typespec for possible event type"

        def fork_it_over(self) -> int:
            ...

    did = FileListDataset("/foo/bar/baz.root")
    datasource = FuncADLQuery[my_type_info](
        dataset_identifier=did, codegen="uproot", item_type=my_type_info
    )

    assert datasource.item_type == my_type_info


def test_type_any():
    "Test the type is any if no type is given"
    did = FileListDataset("/foo/bar/baz.root")
    datasource = FuncADLQuery(dataset_identifier=did, codegen="uproot")
    assert datasource.item_type == Any



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations
from func_adl.object_stream import S

from typing import Union, Callable, Iterable
import ast
from servicex import DatasetGroup

from servicex.func_adl.func_adl_dataset import T


class FuncADLDatasetGroup(DatasetGroup):
    def SelectMany(
        self, func: Union[str, ast.Lambda, Callable[[T], Iterable[S]]]
    ) -> FuncADLDatasetGroup:
        self.datasets = [dataset.SelectMany(func) for dataset in self.datasets]
        return self

    def Select(
        self, func: Union[str, ast.Lambda, Callable[[T], Iterable[S]]]
    ) -> FuncADLDatasetGroup:
        self.datasets = [dataset.Select(func) for dataset in self.datasets]
        return self

    def Where(
        self, func: Union[str, ast.Lambda, Callable[[T], Iterable[S]]]
    ) -> FuncADLDatasetGroup:
        self.datasets = [dataset.Where(func) for dataset in self.datasets]
        return self



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import ast
from typing import Optional, cast


class FuncADLServerException (Exception):
    'Thrown when an exception happens contacting the server'
    def __init__(self, msg):
        Exception.__init__(self, msg)


def has_tuple(a: ast.AST) -> bool:
    """Determine if this query used tuples in its final result

    NOTE: This can't do depth searches in a really complex
    query - then you need to follow best practices.

    Args:
        a (ast.AST): The query

    Returns:
        bool: True if the final Select statement has tuples or not.
    """

    def find_Select(a: ast.AST) -> Optional[ast.Call]:
        # Descent the call chain until we find a Select.
        while isinstance(a, ast.Call):
            if isinstance(a.func, ast.Name):
                if cast(ast.Name, a.func).id == "Select":
                    return a
            a = a.args[0]
        return None

    select_statement = find_Select(a)
    if select_statement is None:
        return False

    # Ok - we have a select statement. Now we need to see if it is returning a tuple.
    func_called = select_statement.args[1]
    assert isinstance(func_called, ast.Lambda)
    body = func_called.body
    return isinstance(body, ast.Tuple)


def has_col_names(a: ast.AST) -> bool:
    """Determine if any column names were specified
    in this request.

    Args:
        a (ast.AST): The complete AST of the request.

    Returns:
        bool: True if no column names were specified, False otherwise.
    """
    assert isinstance(a, ast.Call)
    func_ast = a
    top_function = cast(ast.Name, a.func).id

    if top_function == "ResultAwkwardArray":
        if len(a.args) >= 2:
            cols = a.args[1]
            if isinstance(cols, ast.List):
                if len(cols.elts) > 0:
                    return True
            elif isinstance(ast.literal_eval(cols), str):
                return True
        func_ast = a.args[0]
        assert isinstance(func_ast, ast.Call)

    top_function = cast(ast.Name, func_ast.func).id
    if top_function not in ["Select", "SelectMany"]:
        return False

    # Grab the lambda and see if it is returning a dict
    func_called = func_ast.args[1]
    assert isinstance(func_called, ast.Lambda)
    body = func_called.body
    if isinstance(body, ast.Dict):
        return True

    # Ok - we didn't find evidence of column names being
    # specified. It could still happen, but not as far
    # as we can tell.

    return False



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
import os
from pathlib import Path
from typing import List, Optional

from tinydb import TinyDB, Query, where

from servicex.configuration import Configuration
from servicex.models import TransformRequest, TransformStatus, TransformedResults


class CacheException(Exception):
    pass


class QueryCache:
    def __init__(self, config: Configuration):
        self.config = config
        Path(self.config.cache_path).mkdir(parents=True, exist_ok=True)
        self.db = TinyDB(os.path.join(self.config.cache_path, "db.json"))

    def close(self):
        self.db.close()

    def cache_transform(self, transform: TransformRequest,
                        completed_status: TransformStatus, data_dir: str,
                        file_list: List[str],
                        signed_urls) -> TransformedResults:

        record = TransformedResults(
            hash=transform.compute_hash(),
            title=transform.title,
            codegen=transform.codegen,
            request_id=completed_status.request_id,
            submit_time=completed_status.submit_time,
            data_dir=data_dir,
            file_list=file_list,
            signed_url_list=signed_urls,
            files=completed_status.files,
            result_format=transform.result_format
        )
        self.db.insert(json.loads(record.json()))
        return record

    def update_record(self, record: TransformedResults):
        transforms = Query()
        self.db.update(json.loads(record.json()), transforms.hash == record.hash)

    def get_transform_by_hash(self, hash: str) -> Optional[TransformedResults]:
        transforms = Query()
        records = self.db.search(transforms.hash == hash)
        if not records:
            return None

        if len(records) != 1:
            raise CacheException("Multiple records found in db for hash")
        else:
            return TransformedResults(**records[0])

    def get_transform_by_request_id(self, request_id: str) -> Optional[TransformedResults]:
        transforms = Query()
        records = self.db.search(transforms.request_id == request_id)
        if not records:
            return None

        if len(records) != 1:
            raise CacheException("Multiple records found in db for request_id")
        else:
            return TransformedResults(**records[0])

    def cache_path_for_transform(self, transform_status: TransformStatus) -> Path:
        base = Path(self.config.cache_path)
        result = Path(os.path.join(base, transform_status.request_id))
        result.mkdir(parents=True, exist_ok=True)
        return result

    def cached_queries(self) -> List[TransformedResults]:
        transforms = Query()
        return [TransformedResults(**doc) for doc in
                self.db.search(transforms.request_id.exists())]

    def delete_record_by_request_id(self, request_id: str):
        self.db.remove(where('request_id') == request_id)

    def get_codegen_by_backend(self, backend: str) -> Optional[dict]:
        codegens = Query()
        records = self.db.search(codegens.backend == backend)
        if not records:
            return None

        if len(records) != 1:
            raise CacheException("Multiple records found in db for same backend")
        else:
            return records[0]

    def update_codegen_by_backend(self, backend: str, codegen_list: list) -> Optional[str]:
        transforms = Query()
        self.db.upsert({'backend': backend, 'codegens': codegen_list},
                       transforms.backend == backend)

    def delete_codegen_by_backend(self, backend: str):
        self.db.remove(where('backend') == backend)



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os.path
from hashlib import sha1
from pathlib import Path
from typing import List

from miniopy_async import Minio

from servicex.models import ResultFile, TransformStatus


def _sanitize_filename(fname: str):
    "No matter the string given, make it an acceptable filename on all platforms"
    return fname.replace("*", "_").replace(";", "_").replace(":", "_")


class MinioAdapter:
    # This must be at least 40, the length of the `hash` we are using, or
    # undefined things will happen.
    MAX_PATH_LEN = 60

    def __init__(
        self,
        endpoint_host: str,
        secure: bool,
        access_key: str,
        secret_key: str,
        bucket: str,
    ):
        self.minio = Minio(
            endpoint_host, access_key=access_key, secret_key=secret_key, secure=secure
        )

        self.bucket = bucket

    @classmethod
    def for_transform(cls, transform: TransformStatus):
        return MinioAdapter(
            endpoint_host=transform.minio_endpoint,
            secure=transform.minio_secured,
            access_key=transform.minio_access_key,
            secret_key=transform.minio_secret_key,
            bucket=transform.request_id,
        )

    async def list_bucket(self) -> List[ResultFile]:
        objects = await self.minio.list_objects(self.bucket)
        return [
            ResultFile(
                filename=obj.object_name,
                size=obj.size,
                extension=obj.object_name.split(".")[-1],
            )
            for obj in objects
        ]

    async def download_file(
        self, object_name: str, local_dir: str, shorten_filename: bool = False
    ) -> Path:
        os.makedirs(local_dir, exist_ok=True)
        path = Path(
            os.path.join(
                local_dir,
                _sanitize_filename(
                    self.hash_path(object_name) if shorten_filename else object_name,
                ),
            )
        )

        _ = await self.minio.fget_object(
            bucket_name=self.bucket, object_name=object_name, file_path=path.as_posix()
        )
        return path.resolve()

    async def get_signed_url(self, object_name: str) -> str:
        return await self.minio.get_presigned_url(
            bucket_name=self.bucket, object_name=object_name, method="GET"
        )

    @classmethod
    def hash_path(cls, file_name):
        """
        Make the path safe for object store or POSIX, by keeping the length
        less than MAX_PATH_LEN. Replace the leading (less interesting) characters with a
        forty character hash.
        :param file_name: Input filename
        :return: Safe path string
        """
        if len(file_name) > cls.MAX_PATH_LEN:
            hash = sha1(file_name.encode("utf-8")).hexdigest()
            return "".join(
                [
                    "_",
                    hash,
                    file_name[-1 * (cls.MAX_PATH_LEN - len(hash) - 1) :],  # noqa: E203
                ]
            )
        else:
            return file_name



# Copyright (c) 2023, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import yaml
import pathlib
from typing import Any, Dict, Union
import rich


def load_databinder_config(input_config:
                           Union[str, pathlib.Path, Dict[str, Any]]
                           ) -> Dict[str, Any]:
    """
    Loads, validates, and returns DataBinder configuration.
    The order of function call matters.
    Args:
        input_config (Union[str, pathlib.Path, Dict[str, Any]]):
            path to config file or config as dict
    Returns:
        Dict[str, Any]: configuration
    """
    def prepare_config(config):
        _set_default_values(config)
        ndef = _replace_definition_in_sample_block(config)
        rich.print(f"Replaced {ndef} Option values from Definition block")
        _support_old_option_names(config)
        _decorate_backend_per_sample(config)
        return _validate_config(config)

    if isinstance(input_config, Dict):
        return prepare_config(input_config)
    else:
        file_path = pathlib.Path(input_config)
        config = yaml.safe_load(file_path.read_text())
        rich.print(f"Loading DataBinder config file: {file_path}")
        return prepare_config(config)


def _set_default_values(config: Dict[str, Any]) -> Dict:
    """
    Default values in General block:
        Delivery: LocalPath
        OutputFormat: root
    """
    if 'Delivery' in config['General'].keys():
        config['General']['Delivery'] = config['General']['Delivery'].lower()
    else:
        config['General']['Delivery'] = 'localpath'

    if 'OutputFormat' not in config['General'].keys():
        config['General']['OutputFormat'] = 'root'

    if 'OutFilesetName' not in config['General'].keys():
        config['General']['OutFilesetName'] = 'servicex_fileset'

    for (idx, sample) in zip(range(len(config['Sample'])), config['Sample']):
        if 'IgnoreLocalCache' not in sample.keys():
            if 'IgnoreLocalCache' in config['General'].keys():
                config['Sample'][idx]['IgnoreLocalCache'] = config['General']['IgnoreLocalCache']
            else:
                config['Sample'][idx]['IgnoreLocalCache'] = False

    return config


def _replace_definition_in_sample_block(config: Dict[str, Any]):
    """
    Replace DEF_X in the Sample block with
    a value of the same DEF_X key in the Definition block
    """
    ndef = 0
    definition = config.get('Definition')
    if definition is not None:
        samples = config.get('Sample')
        if samples is not None:
            for n, sample in enumerate(samples):
                for field, value in sample.items():
                    if isinstance(value, str):
                        if 'DEF_' in value:
                            def_in_value = True
                            for repre, new_str in definition.items():
                                if repre == value:
                                    samples[n][field] \
                                        = samples[n][field] \
                                        .replace(repre, new_str)
                                    ndef = ndef + 1
                                    def_in_value = False
                            if def_in_value:
                                raise NotImplementedError(
                                    f"{value} is NOT defined in the Definition block"
                                )
        return ndef
    else:
        return ndef


def _support_old_option_names(config: Dict[str, Any]) -> Dict:
    """ """
    if 'ServiceXName' in config['General'].keys():
        config['General']['ServiceX'] = config['General'].pop('ServiceXName')

    if 'Transformer' in config['General'].keys():
        config['General']['Codegen'] = config['General'].pop('Transformer')
    for (idx, sample) in zip(range(len(config['Sample'])), config['Sample']):
        if 'Transformer' in sample.keys():
            config['Sample'][idx]['Codegen'] = config['Sample'][idx].pop('Transformer')

    if 'IgnoreServiceXCache' in config['General'].keys():
        config['General']['IgnoreLocalCache'] = config['General'].pop('IgnoreServiceXCache')
    for (idx, sample) in zip(range(len(config['Sample'])), config['Sample']):
        if 'IgnoreServiceXCache' in sample.keys():
            config['Sample'][idx]['IgnoreLocalCache'] = config['Sample'][idx].pop(
                'IgnoreServiceXCache')
    return config


def _decorate_backend_per_sample(config: Dict[str, Any]) -> Dict:
    """ from General block """
    pair = ("default_transformer", "default_codegen")
    if 'Codegen' in config['General'].keys():
        if config['General']['Codegen'] == "atlasr21":
            pair = ("xaod", "atlasr21")
        elif config['General']['Codegen'] == "uproot":
            pair = ("uproot", "uproot")
        elif config['General']['Codegen'] == "python":
            pair = ("uproot", "python")

    """ from Sample block """
    for (idx, sample) in zip(range(len(config['Sample'])), config['Sample']):
        if 'Codegen' in sample.keys():
            if sample['Codegen'] == "atlasr21":
                config['Sample'][idx]['Transformer'] = "xaod"
            elif sample['Codegen'] == "uproot":
                config['Sample'][idx]['Transformer'] = "uproot"
            elif sample['Codegen'] == "python":
                config['Sample'][idx]['Transformer'] = "uproot"
        else:
            config['Sample'][idx]['Transformer'] = pair[0]
            config['Sample'][idx]['Codegen'] = pair[1]

    return config


def _validate_config(config: Dict[str, Any]):
    """Returns True if the config file is validated,
    otherwise raises exceptions.
    Checks that the config satisfies the json schema,
    and performs additional checks to
    validate the config further.
    Args:
        config (Dict[str, Any]): configuration
    Raises:
        NotImplementedError
        ValueError
        KeyError
    Returns:
        bool: whether the validation was successful
    """

    # Option names
    available_keys = [
        'General', 'ServiceX', 'OutputDirectory', 'Transformer', 'Codegen',
        'OutputFormat', 'OutFilesetName', 'IgnoreLocalCache', 'Delivery',
        'Name',
        'Sample', 'RucioDID', 'XRootDFiles', 'Tree',
        'Filter', 'Columns', 'FuncADL', 'LocalPath', 'Definition',
        'Function', 'NFiles'
    ]

    # General and Sample are mandatory blocks
    if 'General' not in config.keys() and 'Sample' not in config.keys():
        raise KeyError("You should have 'General' block and "
                       "at least one 'Sample' block in the config")

    # Check all Opion names
    keys_in_config = set()
    for item in config['General']:
        keys_in_config.add(item)
    for sample in config['Sample']:
        for item in sample.keys():
            keys_in_config.add(item)
    for key in keys_in_config:
        if key not in available_keys:
            raise KeyError(f"Unknown Option {key} in the config")

    # Check General block option values
    if 'Delivery' in config['General'].keys():
        if config['General']['Delivery'] not in [
                'localpath', 'localcache', 'objectstore']:
            raise ValueError(
                f"Unsupported delivery option: {config['General']['Delivery']}"
                f" - supported options: LocalPath, LocalCache, ObjectStore"
            )
    if 'ServiceX' not in config['General'].keys():
        raise KeyError(
            "Option 'ServiceXName' is required in General block"
        )
    if config['General']['OutputFormat'].lower() != 'parquet' and \
            config['General']['OutputFormat'].lower() != 'root':
        raise ValueError(
            "OutputFormat can be either parquet or root"
        )

    # Check Sample block option values
    for sample in config['Sample']:
        if ('RucioDID' not in sample.keys()) \
                and ('XRootDFiles' not in sample.keys()) \
                and ('LocalPath' not in sample.keys()):
            raise KeyError(
                "Please specify a valid input source "
                f"for Sample {sample['Name']} e.g. RucioDID, XRootDFiles"
            )
        if 'RucioDID' in sample.keys():
            for did in sample['RucioDID'].split(","):
                if len(did.split(":")) != 2:
                    raise ValueError(
                        f"Sample {sample['Name']} "
                        f"- RucioDID {did} is missing the scope"
                    )
        if ('Tree' in sample) and \
                ('uproot' not in sample['Transformer']):
            raise KeyError(
                f"Option Tree in Sample {sample['Name']} "
                "is only available for uproot transformer"
            )
        if 'Columns' in sample and 'FuncADL' in sample:
            raise KeyError(
                f"Sample {sample['Name']} - Use one type of query per sample: "
                "Columns for TCut and FuncADL for func-adl"
            )
        if 'FuncADL' in sample and 'Filter' in sample:
            raise KeyError(
                f"Sample {sample['Name']} - "
                "You cannot use Filter with func-adl query"
            )

    return config



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import asyncio
from pathlib import Path
from typing import Optional, List

import rich
import typer
from rich.progress import Progress
from rich.table import Table

from servicex.app.cli_options import url_cli_option, backend_cli_option
from servicex.minio_adapter import MinioAdapter
from servicex.models import Status, ResultFile
from servicex.servicex_client import ServiceXClient

transforms_app = typer.Typer(name="transforms", no_args_is_help=True)


@transforms_app.callback()
def transforms():
    """
    Sub-commands for interacting with transforms.
    """
    pass


@transforms_app.command(no_args_is_help=True)
def list(
    url: Optional[str] = url_cli_option,
    backend: Optional[str] = backend_cli_option,
    complete: Optional[bool] = typer.Option(
        None, "--complete", help="Only show successfully completed transforms"
    ),
):
    """
    List the transforms that have been run.
    """
    sx = ServiceXClient(url=url, backend=backend)
    table = Table(title="ServiceX Transforms")
    table.add_column("Transform ID")
    table.add_column("Title")
    table.add_column("Status")
    table.add_column("Files")
    transforms = sx.get_transforms()
    for t in transforms:
        if not complete or complete and t.status == Status.complete:
            table.add_row(
                t.request_id, "Not implemented", t.status, str(t.files_completed)
            )

    rich.print(table)


@transforms_app.command(no_args_is_help=True)
def files(
    url: Optional[str] = url_cli_option,
    backend: Optional[str] = backend_cli_option,
    transform_id: str = typer.Option(None, "-t", "--transform-id", help="Transform ID"),
):
    """
    List the files that were produced by a transform.
    """
    async def list_files(sx: ServiceXClient, transform_id: str) -> List[ResultFile]:
        transform = await sx.get_transform_status_async(transform_id)
        minio = MinioAdapter.for_transform(transform)
        return await minio.list_bucket()

    sx = ServiceXClient(url=url, backend=backend)
    result_files = asyncio.run(list_files(sx, transform_id))
    table = rich.table.Table(title=f"Files from {transform_id}")
    table.add_column("filename")
    table.add_column("Size(Mb)")
    table.add_column("Filetype")
    for f in result_files:
        table.add_row(f.filename, "%.2f" % (f.size / 1e6), f.extension)
    rich.print(table)


@transforms_app.command(no_args_is_help=True)
def download(
    url: Optional[str] = url_cli_option,
    backend: Optional[str] = backend_cli_option,
    transform_id: str = typer.Option(None, "-t", "--transform-id", help="Transform ID"),
    local_dir: str = typer.Option(".", "-d", help="Local dir to download to"),
):
    """
    Download the files that were produced by a transform.
    """
    async def download_files(sx: ServiceXClient, transform_id: str, local_dir):
        async def download_with_progress(filename) -> Path:
            p = await minio.download_file(
                filename,
                local_dir,
                shorten_filename=sx.config.shortened_downloaded_filename,
            )
            progress.advance(download_progress)
            return p

        transform = await sx.get_transform_status_async(transform_id)
        minio = MinioAdapter.for_transform(transform)
        file_list = await minio.list_bucket()
        progress.update(download_progress, total=len(file_list))
        progress.start_task(download_progress)

        tasks = [download_with_progress(f.filename) for f in file_list]
        return await asyncio.gather(*tasks)

    with Progress() as progress:
        download_progress = progress.add_task("Downloading", start=False, total=None)
        sx = ServiceXClient(url=url, backend=backend)
        result_files = asyncio.run(download_files(sx, transform_id, local_dir))

    for path in result_files:
        print(path.as_posix())



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import hashlib
from datetime import datetime
from enum import Enum

from pydantic import BaseModel, Field, validator
from typing import List, Optional


class ResultDestination(str, Enum):
    r"""
    Direct the output to object store or posix volume
    """
    object_store = "object-store"
    volume = "volume"


class ResultFormat(str, Enum):
    r"""
    Specify the file format for the generated output
    """
    parquet = "parquet"
    root = "root-file"


class Status(str, Enum):
    r"""
    Status of a submitted transform
    """
    complete = ("Complete",)
    fatal = ("Fatal",)
    canceled = ("Canceled",)
    submitted = ("Submitted",)
    looking = ("Lookup",)
    running = "Running"


class TransformRequest(BaseModel):
    r"""
    Transform request sent to ServiceX
    """
    title: Optional[str] = None
    did: Optional[str] = None
    file_list: Optional[List[str]] = Field(default=None, alias="file-list")
    selection: str
    image: Optional[str] = None
    codegen: str
    tree_name: Optional[str] = Field(default=None, alias="tree-name")
    result_destination: ResultDestination = Field(alias="result-destination")
    result_format: ResultFormat = Field(alias="result-format")

    class Config:
        allow_population_by_field_name = True

    def compute_hash(self):
        r"""
        Compute a hash for this submission. Only include properties that impact the result
        so we have maximal ability to reuse transforms

        :return: SHA256 hash of request
        """
        sha = hashlib.sha256(
            str(
                [
                    self.did,
                    self.selection,
                    self.tree_name,
                    self.codegen,
                    self.image,
                    self.result_format.name,
                    self.file_list,
                ]
            ).encode("utf-8")
        )
        return sha.hexdigest()


class TransformStatus(BaseModel):
    r"""
    Status object returned by servicex
    """
    request_id: str
    did: str
    selection: str
    tree_name: Optional[str] = Field(alias="tree-name")
    image: str
    result_destination: ResultDestination = Field(alias="result-destination")
    result_format: ResultFormat = Field(alias="result-format")
    generated_code_cm: str = Field(alias="generated-code-cm")
    status: Status
    app_version: str = Field(alias="app-version")
    files: int
    files_completed: int = Field(alias="files-completed")
    files_failed: int = Field(alias="files-failed")
    files_remaining: Optional[int] = Field(alias="files-remaining")
    submit_time: datetime = Field(alias="submit-time")
    finish_time: Optional[datetime] = Field(alias="finish-time")
    minio_endpoint: Optional[str] = Field(alias="minio-endpoint")
    minio_secured: Optional[bool] = Field(alias="minio-secured")
    minio_access_key: Optional[str] = Field(alias="minio-access-key")
    minio_secret_key: Optional[str] = Field(alias="minio-secret-key")

    @validator("finish_time", pre=True)
    def parse_finish_time(cls, v):
        if isinstance(v, str) and v == "None":
            return None
        return v


class ResultFile(BaseModel):
    r"""
    Record reporting the properties of a transformed file result
    """
    filename: str
    size: int
    extension: str


class TransformedResults(BaseModel):
    r"""
    Returned for a submission. Gives you everything you need to know about a completed
    transform.
    """
    hash: str
    title: str
    codegen: str
    request_id: str
    submit_time: datetime
    data_dir: str
    file_list: List[str]
    signed_url_list: List[str]
    files: int
    result_format: ResultFormat



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from typing import Optional

from rich.progress import Progress, TextColumn, BarColumn, MofNCompleteColumn, \
    TimeRemainingColumn, TaskID


class ProgressCounts:
    def __init__(self,
                 description: str,
                 task_id: TaskID,
                 start: Optional[int] = None,
                 total: Optional[int] = None,
                 completed: Optional[int] = None):

        self.description = description
        self.taskId = task_id
        self.start = start
        self.total = total
        self.completed = completed


class ExpandableProgress:
    def __init__(self,
                 display_progress: bool = True,
                 provided_progress: Optional[Progress | ExpandableProgress] = None,
                 overall_progress: bool = False):
        """
        We want to be able to use rich progress bars in the async code, but there are
        some situtations where the user doesn't want them. Also we might be running
        several simultaneous progress bars, and we want to be able to control that.

        We still want to keep the context manager interface, so this class implements
        the context manager but if display_progress is False, then it does nothing.
        If provided_progress is set then we just use that. Otherwise we create a new
        progress bar

        :param display_progress:
        :param provided_progress:
        """
        self.display_progress = display_progress
        self.provided_progress = provided_progress
        self.overall_progress = overall_progress
        self.overall_progress_transform_task = None
        self.overall_progress_download_task = None
        self.progress_counts = {}
        if display_progress:
            if self.overall_progress or not provided_progress:
                self.progress = Progress(
                    TextColumn("[progress.description]{task.description}"),
                    BarColumn(),
                    MofNCompleteColumn(),
                    TimeRemainingColumn(compact=True, elapsed_when_finished=True)
                )

            if provided_progress:
                self.progress = provided_progress if isinstance(provided_progress, Progress) \
                    else provided_progress.progress
        else:
            self.progress = None

    def __enter__(self):
        """
        Start the progress bar if it is not already started and the user wants one.
        :return:
        """
        if self.display_progress and not self.provided_progress:
            self.progress.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Close the progress bar if it is not already closed and the user wanted one in
        the first place.
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        if self.display_progress and not self.provided_progress:
            self.progress.stop()

    def add_task(self, param, start, total):
        if self.display_progress and self.overall_progress:
            if (
                not self.overall_progress_download_task
                and not self.overall_progress_transform_task
            ):
                self.overall_progress_transform_task = self.progress.add_task("Transform",
                                                                              start=False,
                                                                              total=None)
                self.overall_progress_download_task = self.progress.add_task("Download/URLs",
                                                                             start=False,
                                                                             total=None)

            task_id = self.progress.add_task(param, start=start, total=total, visible=False)
            new_task = ProgressCounts(param, task_id, start=start, total=total)
            self.progress_counts[task_id] = new_task
            return task_id
        if self.display_progress and not self.overall_progress:
            return self.progress.add_task(param, start=start, total=total)

    def update(self, task_id, task_type, total=None, completed=None):

        if self.display_progress and self.overall_progress:
            # Calculate and update
            overall_completed = 0
            overall_total = 0
            if completed:
                self.progress_counts[task_id].completed = completed

            elif total:
                self.progress_counts[task_id].total = total

            for task in self.progress_counts:
                if (
                    self.progress_counts[task].description == task_type
                    and self.progress_counts[task].completed
                ):
                    overall_completed += self.progress_counts[task].completed

            for task in self.progress_counts:
                if (
                    self.progress_counts[task].description == task_type
                    and self.progress_counts[task].total
                ):
                    overall_total += self.progress_counts[task].total

            if task_type == "Transform":
                return self.progress.update(self.overall_progress_transform_task,
                                            completed=overall_completed,
                                            total=overall_total)
            else:
                return self.progress.update(self.overall_progress_download_task,
                                            completed=overall_completed,
                                            total=overall_total)

        if self.display_progress and not self.overall_progress:
            return self.progress.update(task_id, completed=completed, total=total)

    def start_task(self, task_id, task_type):
        if self.display_progress and self.overall_progress:
            if task_type == "Transform":
                self.progress.start_task(task_id=self.overall_progress_transform_task)
            else:
                self.progress.start_task(task_id=self.overall_progress_download_task)
        elif self.display_progress and not self.overall_progress:
            self.progress.start_task(task_id=task_id)

    def advance(self, task_id, task_type):
        if self.display_progress and self.overall_progress:
            if task_type == "Transform":
                self.progress.advance(task_id=self.overall_progress_transform_task)
            else:
                self.progress.advance(task_id=self.overall_progress_download_task)
        elif self.display_progress and not self.overall_progress:
            self.progress.advance(task_id=task_id)
